#include <errno.h>
#include <time.h>
#include <limits.h>
#include <math.h>

#include <unistd.h>
#include <fcntl.h>
#include <spawn.h>
#include <signal.h>
#include <poll.h>
#include <sys/socket.h>
#include <err.h>

#include "libreap.h"


extern char **environ;


static inline int recvmsg_with_fd(int sock, void *buf, size_t buflen, int flags, int *fd)
{
	union {
		struct cmsghdr cmsghdr;
		char control[CMSG_SPACE(sizeof(int))];
	} cmsg = { 0 };
	struct msghdr msg = {
		.msg_iov = &(struct iovec) {
			.iov_base = buf,
			.iov_len = buflen,
		},
		.msg_iovlen = 1,
		.msg_control = &cmsg,
		.msg_controllen = sizeof(cmsg),
	};
	int r = recvmsg(sock, &msg, flags);
	if (r > 0) {
		*fd = -1;
		for (struct cmsghdr *cp = CMSG_FIRSTHDR(&msg); cp; cp = CMSG_NXTHDR(&msg, cp)) {
			if (cp->cmsg_level != SOL_SOCKET)
				continue;
			if (cp->cmsg_type != SCM_RIGHTS)
				continue;
			if (cp->cmsg_len != CMSG_LEN(sizeof(int)))
				continue;
			*fd = *(int *)CMSG_DATA(cp);
			(void)fcntl(*fd, F_SETFD, FD_CLOEXEC);
		}
	}
	return r;
}

static inline int block_signal(int sig)
{
	sigset_t sigs;
	sigfillset(&sigs);
	struct sigaction act = {
		.sa_handler = SIG_IGN,
		.sa_mask = sigs,
	};
	return sigaction(sig, &act, NULL) < 0 ? -errno : 0;
}



#define SM_REAPER_SPAWN_DELAY_BASE_MS 100
#define SM_REAPER_SPAWN_DELAY_MULT    1.2
#define SM_REAPER_SPAWN_DELAY_MAX_MS  5000

static pid_t spawn(char **argv, int sock)
{
	pid_t p = (pid_t)-1;

	sigset_t ss;
	sigfillset(&ss);

	posix_spawnattr_t at;
	posix_spawnattr_init(&at);
	posix_spawnattr_setsigdefault(&at, &ss);
	posix_spawnattr_setflags(&at, POSIX_SPAWN_SETSIGDEF);

	posix_spawn_file_actions_t fa;
	posix_spawn_file_actions_init(&fa);
	posix_spawn_file_actions_adddup2(&fa, sock, SM_REAPER_FD);

	int r;
	unsigned long delay_ms = SM_REAPER_SPAWN_DELAY_BASE_MS;
	while ((r = posix_spawnp(&p, argv[0], &fa, &at, argv, environ)) != 0) {
		if (r == EINVAL)
			break;

		warnc(r, "could not spawn child, waiting %lu ms", delay_ms);
		struct timespec delay = {
			.tv_sec = delay_ms / 1000,
			.tv_nsec = (delay_ms % 1000) * 1000000,
		};
		while (nanosleep(&delay, &delay) < 0 && errno == EINTR) {}

		if (delay_ms >= SM_REAPER_SPAWN_DELAY_MAX_MS)
			delay_ms = SM_REAPER_SPAWN_DELAY_MAX_MS;
		else
			delay_ms *= SM_REAPER_SPAWN_DELAY_MULT;
	}
	return p;
}

void sm_reaper_usage(int requested)
{
	errx(requested ? 0 : 255, "usage: PROG [ARG...]");
}

int sm_reaper_main(int argc, char **argv)
{
	if (argc < 2)
		sm_reaper_usage(0);

	int r;
	if ((r = block_signal(SIGPIPE)) < 0)
		warnc(-r, "could not block SIGPIPE");

	pid_t pgid = setsid();
	if (pgid == (pid_t)-1)
		warn("could not start session");

	int reapfd = sm_reaper_open();
	if (reapfd < 0)
		errc(255, -reapfd, "could not open reap fd");

	int procfd[2];
	if (socketpair(AF_LOCAL, SOCK_SEQPACKET, 0, procfd) < 0) {
		/* some systems do not support SOCK_SEQPACKET at all */
		if (socketpair(AF_LOCAL, SOCK_STREAM, 0, procfd) < 0)
			err(255, "could not open process socket");
	}
	(void)fcntl(procfd[0], F_SETFD, FD_CLOEXEC);
	(void)fcntl(procfd[1], F_SETFD, FD_CLOEXEC);

	pid_t childpid = (pid_t)-1;
	struct pollfd polls[] = { 
		{ .fd = reapfd, .events = POLLIN | POLLPRI },
		{ .fd = procfd[0], .events = POLLIN | POLLPRI },
	};
	struct pollfd *reappoll = &polls[0], *procpoll = &polls[1];
	for (;;) {
		if (childpid < 0) {
			childpid = spawn(argv + 1, procfd[1]);
			(void)sm_reaper_add(reappoll->fd, &(struct sm_reaper_proc) {
				.pid = childpid,
				.fd = -1,
			});
		}
		if (poll(polls, 2, -1) < 0)
			continue;

		/* did old processes die? */
		if (reappoll->revents & (POLLIN | POLLPRI)) {
			struct sm_reaped_proc deadproc;
			if (sm_reaper_get(reappoll->fd, &deadproc) > 0) {
				warnx("died: %ld (status: %d, code: %d)", (long)deadproc.pid, deadproc.status, deadproc.code);
				if (deadproc.pid == childpid)
					childpid = (pid_t)-1;
				else
					(void)send(procpoll->fd, &deadproc, sizeof(deadproc), MSG_EOR | MSG_NOSIGNAL);
			}
		}
		/* were new processes spawned? */
		if (procpoll->revents & (POLLIN | POLLPRI)) {
			pid_t pid;
			int pfd = -1;
			int r = recvmsg_with_fd(procpoll->fd, &pid, sizeof(pid), MSG_DONTWAIT, &pfd);
			if (r > 0) {
				(void)sm_reaper_add(reappoll->fd, &(struct sm_reaper_proc) {
					.pid = pid,
					.fd = pfd,
				});
			}
		}
	}

	return 0;
}
