#include <stdatomic.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>

#include <signal.h>
#include <sys/wait.h>
#include "libreap.h"
#include "reaper-internal.h"

static pid_t reaper_pid;
static atomic_int reaper_fd = -1;

static void sm_reap(int sig)
{
	for (;;) {
		struct sm_reaped_proc p = { 0 };
		int stat;
		p.pid = waitpid((pid_t)-1, &stat, WNOHANG);
		if (p.pid == (pid_t)-1) {
			if (errno == EINTR)
				continue;
			break;
		}
		if (p.pid == (pid_t)0)
			break;

		if (WIFEXITED(stat)) {
			p.status = CLD_EXITED;
			p.code = WEXITSTATUS(stat);
		} else if (WIFSIGNALED(stat)) {
			p.status = WCOREDUMP(stat) ? CLD_DUMPED : CLD_KILLED;
			p.code = WTERMSIG(stat);
		} else
			continue;

		int fd = atomic_load(&reaper_fd);
		if (fd < 0)
			break;
		while (write(fd, &p, sizeof(p)) < 0 && errno == EINTR)
			;
	}
}

int sm_reaper_open(void)
{
	/* already reaping? */
	pid_t pid = getpid();
	if (reaper_pid == pid)
		return -EBUSY;

	int fds[2];
	if (pipe(fds) < 0)
		return -errno;
	(void)fcntl(fds[0], F_SETFD, FD_CLOEXEC);
	(void)fcntl(fds[1], F_SETFD, FD_CLOEXEC);

	reaper_pid = pid;
	atomic_store(&reaper_fd, fds[1]);

	sigset_t sigs;
	sigemptyset(&sigs);
	struct sigaction act = {
		.sa_handler = sm_reap,
		.sa_mask    = sigs,
		.sa_flags   = SA_NOCLDSTOP,
	};
	if (sigaction(SIGCHLD, &act, NULL) < 0) {
		int e = errno;
		sm_reaper_close(fds[0]);
		return -e;
	}

	/* set as child subreaper */
	(void)sm_reaper_set_child_reaper(1);
	return fds[0];
}

int sm_reaper_add(int fd, const struct sm_reaper_proc *proc)
{
	/* not needed */
	return 0;
}

int sm_reaper_get(int fd, struct sm_reaped_proc *proc)
{
	int r = read(fd, proc, sizeof(*proc));
	if (r < 0)
		return -errno;
	return r > 0 ? 1 : 0;
}

void sm_reaper_close(int fd)
{
	if (reaper_pid != getpid())
		return;

	(void)sm_reaper_set_child_reaper(0);

	sigset_t sigs;
	sigfillset(&sigs);
	struct sigaction act = {
		.sa_handler = SIG_DFL,
		.sa_mask = sigs,
	};
	(void)sigaction(SIGCHLD, &act, NULL);

	(void)close(fd);
	(void)close(reaper_fd);
	atomic_store(&reaper_fd, -1);
	reaper_pid = 0;
}
